#include <stdlib.h>
#include <string.h>
#include "GrdAPI.h"

#include "writemask.h"

#define    VERSION                  3.0427

// Dongle memory parameters
#define		PIAT_HEADER				6		// Header of PIAT (Protected Item Allocation Table)
#define		PIAT_RESERVED			2		// Number of reserved bytes at the end of PIAT
#define		AST_HEADER				18		// Header of AST (Algorithm System Table), Sign/Time or higher dongles
#define		AST_RESERVED			8		// Number of reserved bytes at the end of AST
#define		MAX_ITEM_HEADER			76		// Size of Item Header in case last generation dongle is being used
#define		GrdWM_SAM_Offset		44		// SAM offset of user memory
#define		GrdWM_UAM_Offset		14		// UAM offset of user memory
#define		DONGLE_RESERVED			8
#define		MAX_LMS_DET_SIZE		254		// Max length of License Table descriptor
#define		MAX_CODE_DET_SIZE		255		// Max length of User Firmware descriptor
#define		OLD_DET_LIMIT			255		// Previous (not Sign/Time) dongle's determinant limit

#define		UnknownDongle			0
#define		StealthII				(DWORD)1<<0
#define		StealthIII				(DWORD)1<<1
#define		SignTime				(DWORD)1<<2

// Guardant Code Dongles memory parameters
#define		CODE_FLASH_START		0x00020000
#define		CODE_FLASH_SIZE			0x00020000
#define		CODE_SECTOR_SIZE		0x00008000
#define		CODE_RAM_START			0x40003000
#define		CODE_RAM_SIZE			0x00004fe0

void CWritemask::CopyData(BYTE *lpvSrc, BYTE *lpvDst, DWORD dwLen)
{
	DWORD i;
	if (lpvSrc && lpvDst)
		for (i = 0; i < dwLen; i++)
			lpvDst[i] = lpvSrc[i];

	if (!lpvSrc && lpvDst)
		for (i = 0; i < dwLen; i++)
			lpvDst[i] = 0;
}

BYTE byNumberOfSectors(BYTE byCodeNum, char *szMaskImage)
{
	BYTE bySecNumber = 0;
	BYTE byCodeNumCounter = 0;
	BYTE i = 0;
	
	while( szMaskImage[i] )
	{
		bySecNumber++;

		if( szMaskImage[i] != szMaskImage[i+1] )
		{
			if( byCodeNumCounter == byCodeNum ) return bySecNumber;
			bySecNumber = 0;
			byCodeNumCounter++;
		}
		i++;
	}

	return 0;
}

BYTE byStartSectorNumber(BYTE byCodeNum, char *szMaskImage)
{
	BYTE byCodeNumCounter = 0;
	BYTE i = 1;
	
	while( szMaskImage[i] )
	{
		if( byCodeNumCounter == byCodeNum )
			return i-1;

		if( szMaskImage[i-1] != szMaskImage[i] )
			byCodeNumCounter++;		

		i++;
	}

	if( byCodeNumCounter == byCodeNum )
		return i-1;
	else
		return 0;
}

void CWritemask::CODE_DET_Prepare(
										BYTE		*pMem,
										BYTE		byNumberOfCode,
										char		*szCodeMaskImage,
										DWORD		dwRAMstart,
										DWORD		dwRAMsize,
										BYTE		*pPubECCKey4Sign,
										BYTE		*pPrivECCKey4Key,
										WORD		*pwCodeDetSize)
{
	DWORD	dwRet = 0;
	DWORD	dwOutLen = 0;
	int		i = 0;

	TGrdLoadableCodeData	*pCODE_struct = (TGrdLoadableCodeData*)pMem;

	// Public part of CODE structure
	pCODE_struct->PublicDataLoadableCode.bLoadableCodeVersion = 0x00;
	pCODE_struct->PublicDataLoadableCode.bReserved0			= 0x00;
	pCODE_struct->PublicDataLoadableCode.bState				= 0x00;
	pCODE_struct->PublicDataLoadableCode.bReserved			= 0x00;
	pCODE_struct->PublicDataLoadableCode.dwLoadingDate		= 0x00000000;

	// Private part of CODE structure
	memcpy(&(pCODE_struct->abLoadableCodePublicKey4VerifySign), pPubECCKey4Sign, GrdECC160_PUBLIC_KEY_SIZE);
	memcpy(&(pCODE_struct->abLoadableCodePrivateKey4DecryptKey), pPrivECCKey4Key, GrdECC160_PRIVATE_KEY_SIZE);

	pCODE_struct->dwBegFlashAddr = CODE_FLASH_START
		+ CODE_SECTOR_SIZE*byStartSectorNumber(byNumberOfCode,szCodeMaskImage);
	pCODE_struct->dwEndFlashAddr = pCODE_struct->dwBegFlashAddr
		+ CODE_SECTOR_SIZE*byNumberOfSectors(byNumberOfCode,szCodeMaskImage) - 1;

	if(dwRAMstart == USE_WHOLE_RAM && dwRAMsize == USE_WHOLE_RAM)
	{
		pCODE_struct->dwBegMemAddr	= CODE_RAM_START;
		pCODE_struct->dwEndMemAddr	= CODE_RAM_START + CODE_RAM_SIZE - 1;
	}
	else
	{
		pCODE_struct->dwBegMemAddr	= dwRAMstart;
		pCODE_struct->dwEndMemAddr	= dwRAMstart + dwRAMsize - 1;
	}

	*pwCodeDetSize = sizeof(TGrdLoadableCodeData);
	return;
}

void CWritemask::LMS_DET_Prepare(	BYTE *pMem, WORD wFlags, BYTE byLMSSize, WORD wLANResource,
										WORD wLMSModules[], WORD *pwLMS_DetSize)
{
	WORD	wCounter = 0;

	switch (CurrentDongle) {

	case StealthII:
		*(BYTE*)(pMem+wCounter) = byLMSSize;					// Number of LMS items
		wCounter++;
		*(BYTE*)(pMem+wCounter) = 0x00;							// Reserved
		wCounter++;

		// Filling LMS table body
		for(BYTE i=0; (i<byLMSSize)&&(i<(BYTE)(sizeof(wLMSModules)/2)); i++)
		{
			*(BYTE*)(pMem+wCounter) = (BYTE)wLMSModules[i];		// 0, 1st, 2nd etc.. module resource of LMS (1 byte)
			wCounter++;
		}
		if ( wCounter%2 != 0 ) {								// Parity byte
			*(BYTE*)(pMem+wCounter) = 0x00;
			wCounter++;
		}
		break;

	default:
		TLMS_FixedHeader lms_header;

		// Preparation of LMS Table header
		lms_header.wSignature = 'LT';					// Signature of LMS Table
		lms_header.wVersion = 0x0510;					// Table's version, defined constant
		lms_header.wFlags = wFlags;						// Flags of LMS Table
		lms_header.byElements = byLMSSize+1;			// Number of records in LMS Table
		memset(lms_header.rs_res, 0, 5);

		memcpy(pMem, &lms_header, sizeof(TLMS_FixedHeader));
		wCounter += sizeof(TLMS_FixedHeader);

		// Filling LMS table body
		if(wFlags & LMS_FLAG_LSIZE_2BYTES)
		{
			*(WORD*)(pMem+wCounter) = wLANResource;				// Real N3 dongle LAN resource (2 bytes)
			wCounter += 2;

			for(BYTE i=0; i < byLMSSize && i < (BYTE)(sizeof(wLMSModules)/2); i++)
			{
				*(WORD*)(pMem+wCounter) = wLMSModules[i];		// 0, 1st, 2nd etc.. module resource of LMS (2 bytes)
				wCounter += 2;
			}
		}
		else
		{
			*(BYTE*)(pMem+wCounter) = (BYTE)wLANResource;		// Real N3 dongle LAN resource (1 byte)
			wCounter++;

			for(BYTE i=0; i < byLMSSize && i < (BYTE)(sizeof(wLMSModules)/2); i++)
			{
				*(BYTE*)(pMem+wCounter) = (BYTE)wLMSModules[i];	// 0, 1st, 2nd etc.. module resource of LMS (1 byte)
				wCounter++;
			}
		}

		*(DWORD*)(pMem+4) = GrdCRC(pMem+8, wCounter-8, Grd_StartCRC);		// Calculate CRC, from Flags(+8) to the end of the table
		break;
	}

	*pwLMS_DetSize = wCounter;

	return;
}

void CWritemask::AST_AddItem(BYTE *pPIAT_AST, WORD wSizeOfNewItem, WORD wName)
{
	WORD wSAM_Offset	= GrdSAM_abyAlgoAddr;						// Offset of User's memory in GrdWM_SAM work mode
	WORD wCurrentElem	= wNumberOfItems;
	WORD wSumItemsSize	= 0;
	WORD wPIATSize		= 0;
	BYTE *pART = NULL;
	BYTE *pAAT = NULL;
	int i = 0;

	switch (CurrentDongle) {

	case StealthII:
		if ( wNumberOfItems == 0 ) {								// There are no Items now
			*(BYTE *)pPIAT_AST = (int)((wSAM_Offset + 8)/2);
		}
		else {														// There are some Items now
			if ( ((wNumberOfItems)+1) > 8 && ((wNumberOfItems)+1)%2 )
				for(i=0; i<(wNumberOfItems); i++)
					*(BYTE *)(pPIAT_AST+i) += 1;

			*(BYTE *)(pPIAT_AST+(wNumberOfItems)) = *(BYTE *)(pPIAT_AST+(wNumberOfItems)-1) + (int)(wSizeOfNewItem/2);
		}

		wPIATSize = (wNumberOfItems+1)<8 ? 8 : (wNumberOfItems+1)+(wNumberOfItems+1)%2;
		break;

	case StealthIII:
		// PIAT (Protected Item Allocation Table)
		wPIATSize = PIAT_HEADER + 2 * (wNumberOfItems) + PIAT_RESERVED;	// Old PIAT Size (Header + Content + Reserved)

		if (wNumberOfItems == 0)									// There are no Items now
		{
			wSumItemsSize = 0;										// Size of the PIAT's content
			*(DWORD *)pPIAT_AST = 0;								// First 4 bytes are 00 00 00 00 (Stealth III)
		}
		else														// There are some Items now
			wSumItemsSize = *(WORD *)(pPIAT_AST + 4) - wPIATSize - wSAM_Offset;	// Calculate old Sum Items Size
		wPIATSize += 2;												// New PIAT Size (record size is 2)
		wSumItemsSize += wSizeOfNewItem;							// New sum size of Items

		*(WORD *)(pPIAT_AST + PIAT_HEADER - 2) = wSAM_Offset + wPIATSize + wSumItemsSize;	// Write down adress of User's memory

		for (i = 0; i < wNumberOfItems; i++)						// Update early created values
			*(WORD *)(pPIAT_AST + PIAT_HEADER + 2*i) += 2;

		*(WORD *)(pPIAT_AST + PIAT_HEADER + 2 * wCurrentElem) = wSAM_Offset + 
			wPIATSize + wSumItemsSize - wSizeOfNewItem;				// New Item adress in PIAT
		break;

	case SignTime:
		// AST (Algorithm System Table). Last generation of dongles
		if(wNumberOfItems == 0)									// There are no Items now
		{
			// Create header
			*(WORD *)(pPIAT_AST) = 0x0100;							// First 2 bytes are 00 (Stealth III) 01 (Sign/Time)
			*(WORD *)(pPIAT_AST + 2) = wSAM_Offset + AST_HEADER + AST_RESERVED;	// Address of User's data (unprotected memory)
			*(WORD *)(pPIAT_AST + 4) = wSAM_Offset + AST_HEADER;	// Address of AAT
			for (i = 0; i < 10; i++)
				*(BYTE *)(pPIAT_AST + 6 + i) = 0;					// Reserved memory (10 bytes)
			*(WORD *)(pPIAT_AST + 16) = wSAM_Offset + AST_HEADER;	// Address of ART
			pART = (BYTE *)malloc(2 * sizeof(BYTE));
			pAAT = (BYTE *)malloc(2 * sizeof(BYTE));
		}
		else {
			pART = (BYTE *)malloc((wNumberOfItems + 1) * 2 * sizeof(BYTE));
			pAAT = (BYTE *)malloc((wNumberOfItems + 1) * 2 * sizeof(BYTE));
			// Take old ART and AAT tables
			CopyData(pPIAT_AST + *(WORD *)(pPIAT_AST + 16) - wSAM_Offset, pART, wNumberOfItems * 2);
			CopyData(pPIAT_AST + *(WORD *)(pPIAT_AST + 4) - wSAM_Offset, pAAT, wNumberOfItems * 2);
		}
		// Update ART and AAT tables
		if (wName > 0xFF00) wName = 0xFF00;
		*(WORD *)(pART + wNumberOfItems * 2) = wName;				// Numeric name of New Item
		for (i = 0; i < wNumberOfItems; i++)
			*(WORD *)(pAAT + i * 2) += 4;							// Update old values of AST
		*(WORD *)(pAAT + wNumberOfItems * 2) = *(WORD *)(pPIAT_AST + 2) + 4;	// Address of New Item (SAM mode!)
		// Update AST header
		*(WORD *)(pPIAT_AST + 2) += (wSizeOfNewItem + 4);			// Update user data address (unprotected data)
		*(WORD *)(pPIAT_AST + 4) += 2;								// Update AAT address
		// Write down updated ART and AAT tables
		CopyData(pART, pPIAT_AST + *(WORD *)(pPIAT_AST + 16) - wSAM_Offset, (wNumberOfItems + 1) * 2);
		CopyData(pAAT, pPIAT_AST + *(WORD *)(pPIAT_AST + 4) - wSAM_Offset, (wNumberOfItems + 1) * 2);
		free(pART);
		free(pAAT);

		wPIATSize = AST_HEADER + 4 * (wNumberOfItems + 1) + AST_RESERVED;
		break;
	}
	(wNumberOfItems)++;											// Inc Number of Items, rutrun
	wAST_Size = wPIATSize;									// Return new wPIATSize
}

void CWritemask::AddAlgorithm(WORD wNumericName, BYTE byLoFlags, WORD wHiFlags, BYTE byAlgorithmCode,
				  WORD wKeyLength, WORD wBlockLength, DWORD dwActivatePwd, DWORD dwDeactivatePwd,
				  DWORD dwReadPwd, DWORD dwUpdatePwd, TGrdTime *pBirthTime, TGrdTime *pDeadTime, TGrdLifeTime *pLifeTime,
				  TGrdFlipTime *pFlipTime, WORD wGP_Counter, WORD wErrorCounter, BYTE *pDet)
{
	BYTE	*pMem;
	WORD	wCounter = 0;											// Offset in Item Header

	if ( CurrentDongle & (StealthIII|SignTime) )
		byLoFlags = byLoFlags | nsafl_ST_III;

	if ( (CurrentDongle & (StealthII|StealthIII) )					// Input correction
		&& ((byAlgorithmCode == rs_algo_GSII64)
		|| (byAlgorithmCode == rs_algo_Sim64Encode)
		|| (byAlgorithmCode == rs_algo_Sim64Decode)
		|| (byAlgorithmCode == rs_algo_HASH64)
		|| (byAlgorithmCode == rs_algo_RND64)))
	{																// Set default algorithm's values
		wBlockLength = 8;
		if (!( (wKeyLength==16) || (wKeyLength==32) ))				// GSII64 determinant length
			wKeyLength = 16;
	}
	if ( !(CurrentDongle & SignTime) ) {							// Determinant length limitations
		if (wKeyLength > OLD_DET_LIMIT)  wKeyLength = OLD_DET_LIMIT;
		if (wBlockLength > OLD_DET_LIMIT)  wBlockLength = OLD_DET_LIMIT;
	}

	if ( byAlgorithmCode == rs_algo_PI )
		wBlockLength = 0;

	if (!pBirthTime)
		wHiFlags &= ~nsafh_BirthTime;
	if (!pDeadTime)
		wHiFlags &= ~nsafh_DeadTime;
	if (!pLifeTime)
		wHiFlags &= ~nsafh_LifeTime;
	if (!pFlipTime)
		wHiFlags &= ~nsafh_FlipTime;

	pMem = (BYTE *)malloc((MAX_ITEM_HEADER + wKeyLength) * sizeof(BYTE));	

	switch (CurrentDongle) {

	case StealthII:
		TStealthII_FixedHeader s2_header;

		s2_header.byLoFlags = byLoFlags;					// LoFlags
		s2_header.byAlgorithmCode = byAlgorithmCode;		// Algorithm code
		s2_header.dwGP_Counter = wGP_Counter;				// GP counter
		s2_header.byKeyLength = (BYTE)wKeyLength;			// Size of algorithm's key							
		s2_header.byBlockLength = (BYTE)wBlockLength;		// Request Block length

		memcpy(pMem, &s2_header, sizeof(TStealthII_FixedHeader));
		wCounter += sizeof(TStealthII_FixedHeader);
		break;

	case StealthIII:
		TStealthIII_FixedHeader s3_header;

		s3_header.byLoFlags = byLoFlags;					// LoFlags
		s3_header.wHiFlags = wHiFlags;						// HiFlags
		s3_header.byAlgorithmCode = byAlgorithmCode;		// Algorithm code
		s3_header.byKeyLength = (BYTE)wKeyLength;			// Size of algorithm's key							
		s3_header.byBlockLength = (BYTE)wBlockLength;		// Request Block length

		memcpy(pMem, &s3_header, sizeof(TStealthIII_FixedHeader));
		wCounter += sizeof(TStealthIII_FixedHeader);

		if (byLoFlags & nsafl_ActivationSrv)
		{
			*(DWORD *)(pMem + wCounter) = dwActivatePwd;	// Activation password, in case flag
			wCounter += 4;									// 'nsafl_ActivationSrv' is set
		}
		if (byLoFlags & nsafl_DeactivationSrv)
		{
			*(DWORD *)(pMem + wCounter) = dwDeactivatePwd;
			wCounter += 4;
		}
		if ((wHiFlags & nsafh_ReadSrv) && (wHiFlags & nsafh_ReadPwd))
		{
			*(DWORD *)(pMem + wCounter) = dwReadPwd;		// Read password, in case flags
			wCounter += 4;									// 'nsafh_ReadSrv' and 'nsafh_ReadPwd' are set
		}
		if (byLoFlags & nsafl_UpdateSrv)
		{
			*(DWORD *)(pMem + wCounter) = dwUpdatePwd;		// Update password, in case flag
			wCounter += 4;									// 'nsafl_UpdateSrv' is set
		}
		if ((byLoFlags & nsafl_GP_dec)
			|| ((byLoFlags & nsafl_GP) && !(byLoFlags & nsafl_ST_III)))
		{
			*(DWORD *)(pMem + wCounter) = wGP_Counter;		// GP counter, in case flag 'nsafl_GP_dec' is set for
			wCounter += 4;									// S/N III or 'nsafl_GP' flag is set for another dongle
		}
		if ((byLoFlags & nsafl_ActivationSrv)
			|| (byLoFlags & nsafl_DeactivationSrv)
			|| ((wHiFlags & nsafh_ReadSrv) && (wHiFlags & nsafh_ReadPwd))
			|| (byLoFlags & nsafl_UpdateSrv))
		{
			*(WORD *)(pMem + wCounter) = wErrorCounter;		// Permissible incorrect password attempts, in case
			wCounter += 2;									// one of the services is enabled
		}
		break;

	case SignTime:											// In case of last generation dongle
		TSignTimeCode_FixedItem item;

		item.byLoFlags = byLoFlags;							// LoFlags
		item.byAlgorithmCode = byAlgorithmCode;				// Algorithm's code
		memset(item.rs_res, 0, 2);							// 2 bytes are reserved for even
		item.dwHiFlags = wHiFlags;							// HiFlags
		item.dwKeyLength = wKeyLength;						// Size of algorithm's key
		item.dwBlockLength = wBlockLength;					// Request Block length
		memset(item.rs_hash, 0, 8);							// 8 bytes are reserved for even
		item.dwActivatePwd = dwActivatePwd;					// Activation service password
		item.dwDeactivatePwd = dwDeactivatePwd;				// Deactivation service password
		item.dwReadPwd = dwReadPwd;							// Read service password
		item.dwUpdatePwd = dwUpdatePwd;						// Update service password

		CopyData((BYTE *)pBirthTime, item.pBirthTime, sizeof(TGrdTime));		// BirthTime struct
		CopyData((BYTE *)pDeadTime, item.pDeadTime, sizeof(TGrdTime));			// DeadTime struct
		CopyData((BYTE *)pLifeTime, item.pLifeTime, sizeof(TGrdLifeTime));		// LifeTime struct
		CopyData((BYTE *)pFlipTime, item.pFlipTime, sizeof(TGrdFlipTime));		// FlipTime struct

		item.dwGP_Counter = wGP_Counter;					// Backward GP counter
		item.dwErrorCounter = wErrorCounter;				// Permissible incorrect password attempts

		memcpy(pMem, &item, sizeof(TSignTimeCode_FixedItem));
		wCounter += sizeof(TSignTimeCode_FixedItem);
		break;
	}

	CopyData((BYTE *)pDet, (BYTE *)(pMem + wCounter), wKeyLength);	// Copy Key into created Item

	CopyData(pMem, pAlgos + wMaskSize, wCounter + wKeyLength);	// Copy Created Algo/Item into image
	free(pMem);
	wMaskSize += (wCounter + wKeyLength);

	AST_AddItem(pAST, wCounter + wKeyLength, wNumericName);
}

void CWritemask::Reinitialize(
		WORD			wDongleType,
		DWORD			dwDongleModel,
		BYTE			*pAlgos_Buffer,
		BYTE			*pAST_buffer)
{
	wNumberOfItems = 0;
	wMaskSize = 0;
	wAST_Size = 0;

	IsInitialized = 1;

	pAlgos = pAlgos_Buffer;
	pAST = pAST_buffer;

	if ( wDongleType & GrdDT_ECC)
		CurrentDongle = SignTime;
	else if ( wDongleType & GrdDT_PI )
		CurrentDongle = StealthIII;
	else if ( wDongleType & GrdDT_GSII64 )
		CurrentDongle = StealthII;
	else
	{
		IsInitialized = 0;
		CurrentDongle = UnknownDongle;
	}
}

CWritemask::CWritemask(
		WORD			wDongleType,
		DWORD			dwDongleModel,
		BYTE			*pAlgos_Buffer,
		BYTE			*pAST_buffer)
{
	Reinitialize(wDongleType, dwDongleModel, pAlgos_Buffer, pAST_buffer);
}